/*    parser.c
 *    Copyright 2002-2004 Kirby Kuehl (vacuum@users.sourceforge.net)
 *    This file is part of pixilate
 *    http://winfingerprint.sourceforge.net/pixilate.php
 *
 *    pixilate is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    pixilate is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *    You should have received a copy of the GNU General Public License
 *    along with pixilate; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "pixilate.h"

/* The PIX 6.2x - 6.3x access list parsing function's main purpose
 * is to tokenize the various parts of an access-list line
 * and hand the tokens to various helper functions such as 
 * port_lookup(), icmp_type_lookup(), protocol_lookup(),
 * operator_port_lookup(), and stats().
 * Upon receiving the data back from these helper functions
 * TCP/UDP/IGMP packets are constructed and written to the wire.
 * ICMP packets are constructed and written to the wire within
 * the icmp_type_lookup() function.
 */

int cisco_parser(char *inputfile, struct acl *pacl, libnet_t *l)
{
    char line[1024];
    int aclid          = 0,
        i              = 0;
    u_long random_int  = 0,
        hostmask       = 0,
        netmask        = 0; /* htonl returns in_addr_t on bsd and solaris  and
                             * u_long on linux */
    u_short range      = 0,
        start_port     = 0,
        end_port       = 0;

    char *token;
    struct in_addr ip, netmaskaddr, start, end, current;
    extern int errno;
    FILE *input;

    /* Open up acl list for parsing */
    if((input = fopen(inputfile, "r")) == NULL)
    {
        fprintf(stderr,"Unable to open %s\n",inputfile);
        exit(EXIT_FAILURE);
    }

    /* Initialization for the entire access list */
    pacl->t_permits = 0;
    pacl->t_denies  = 0;
    pacl->t_errors  = 0;
    srand((unsigned char)time(NULL));

    while(fgets(line,1024,input) != NULL)
    {
        /* Strip carriage returns and line feeds */
        while((line[strlen (line) - 1] == '\r') ||
              (line[strlen (line) - 1] == '\n'))
            line[strlen(line) -1] = '\0';

        /* Initialization for each access list entry */ 
        libnet_clear_packet(l);
        pacl->source_addr      = 0;
        pacl->destination_addr = 0;
        pacl->icmp_type        = 0;
        pacl->rule             = 0;
        pacl->proto_h_len      = 0;
        pacl->source_port      = pacl->def_source_port; 
        pacl->destination_port = pacl->def_destination_port;
        pacl->payload          = pacl->def_payload;

        /* ALL Cisco PIX 6.x access lists begin with 'access-list'    */
        if(strncmp(line,"access-list",11) != 0) /* Not an access-list */
            continue; /* Don't bother to continue parsing             */

        fprintf(stdout,
      "____________________________________________________________________\n");
        fprintf(stdout,"%s\n",line);
      
        /* The first two words of any valid access-list line will be
        * "access-list " followed by an id, so if there is a strtok
        * error the line must have contained only access-list */
        
        if((token = strtok(line,"access-list ")) == NULL)
        {
            fprintf(stdout,"PIXILATE ACL PARSING ERROR: "
                "Could not parse expected acl_ID after \"access-list\"\n");
            pacl->t_errors++;
            continue;
        }

        /* All Cisco PIX 6.x access lists contain an id as their second
         * parameter. Save this parameter for informational purposes in
         * pacl->aclid. This value can be a name or a number so it is left
         * in its original character form for output to stats() then it is
         * passed to atoi() so we can determine if it is a simple (0-99)
         * or extended (100-199) acl. */
        snprintf(pacl->aclid,sizeof(pacl->aclid)-1,"%s",token); 
        aclid = atoi(token);

        /* All Cisco PIX 6.x access lists contain {deny|permit}, 'compiled'
         * or 'remark' as their third parameter, if the parameter is deny 
         * or permit, retain it for informational purposes. If the third
         * parameter is 'compiled', 'remark', or if there is a parsing error,
         * stop parsing the access-list line. */

        if((token = strtok(NULL," ")) == NULL)
        {
            fprintf(stdout, "PIXILATE ACL PARSING ERROR: "
                "Could not parse expected {deny | permit} or compiled\n"); 
            pacl->t_errors++;
            continue;
        }
     
        snprintf(pacl->type,sizeof(pacl->type)-1,"%s",token);
        if(strncmp(pacl->type, "permit",6) == 0)
        {
            if(pacl->options & DENY_ONLY)
               continue; /* Skip all permit acls if -m deny */
            else 
               pacl->t_permits = pacl->t_permits++;
        }
        else if(strncmp(pacl->type, "deny",4) == 0)
        {
            if(pacl->options & PERMIT_ONLY)
               continue; /* Skip all deny acls if -m permit */
            else 
               pacl->t_denies = pacl->t_denies++;
        }
        else if(strncmp(pacl->type, "compiled",8) == 0)
        {
            fprintf(stdout, "PIXILATE NOTE: Skipped TurboACL\n");
            continue;
        }
        else if(strncmp(pacl->type, "remark",6) == 0)
        {
           fprintf(stdout, "PIXILATE NOTE: Skipped remark\n");
           continue;
        }
        else /* should not be reached with a valid access-list */
        {
            fprintf(stderr,"PIXILATE INVALID ACL TYPE: %s\n",token);
            pacl->t_errors++;
            continue;
        }

       /* If processing a simple acl (0-99), we will send a tcp
        * packet, so set token = tcp before sending to protocol_lookup()
        * If processing an extended acl (100-199), use strtok() to obtain
        * the protocol, if there is an error stop parsing line. */
        if(aclid < 99)
            token = "tcp";
        else if((token = strtok(NULL," ")) == NULL)
        {
            fprintf(stdout,"PIXILATE EXTENDED ACL ERROR: "
                "Could not parse expected protocol\n");
            pacl->t_errors++;
            continue;
        }
      
        pacl->protocol = protocol_lookup(token);
        switch(pacl->protocol)
        {
            case 0:
                pacl->rule |= IP_RULE;
                break;
            case 1: 
                pacl->rule |= ICMP_RULE;
                break;
            case 2:
                pacl->rule |= IGMP_RULE;
                pacl->proto_h_len = LIBNET_IGMP_H;
                break;
            case 6:
                pacl->rule |= TCP_RULE;
                pacl->proto_h_len = LIBNET_TCP_H;
                break;
            case 17:
                pacl->rule  |= UDP_RULE;
                pacl->proto_h_len = LIBNET_UDP_H;
                break;
            case 41:
                pacl->proto_h_len = LIBNET_IPV6_H;
                break;
            case 50:
                pacl->rule |= IPSEC_RULE;
                pacl->proto_h_len =
                    LIBNET_IPSEC_ESP_HDR_H + LIBNET_IPSEC_ESP_FTR_H;
                break;
            case 51:
                pacl->rule |= IPSEC_RULE;
                pacl->proto_h_len = LIBNET_IPSEC_AH_H;
                break;
            default: /* We will still attempt to send an ip packet with the 
                      protocol supplied */
                break;
        }

        /* The next value will be the source address for all access-lists
         * (0-99) and (100-199), so if there is an error, stop parsing line. */
        if((token = strtok(NULL," ")) == NULL)
        {

            fprintf(stdout, "PIXILATE ACL PARSING ERROR: "
                "Could not parse source_addr.\n");
            pacl->t_errors++;
            continue;
        }      

        /* Skip ACLs with object-group and obj_group_id */
        if(strncmp(token,"object-group",12) == 0)
        {
            fprintf(stdout,
                "PIXILATE NOTE: Skipped object-group obj_grp_id ACL.\n");
            continue;
        }      

        /* Prepare for source address processing ... */

        if(strncmp(token,"any",3) == 0) /* ACL with source address 'any' */
        {
            if(pacl->def_source_addr == 0)
            {
                /* Get random source address if default source address for 'any'
                 * was not supplied via -s command line option */
                libnet_seed_prand(l);
                pacl->source_addr = libnet_get_prand(LIBNET_PRu32);
            }
            else
                pacl->source_addr = pacl->def_source_addr;
        }
        /* ACL with specific source address defined by 'host' */
        else if(strncmp(token,"host",4) == 0)
        {
            /* 'host' is followed by the source_addr, so skip "host" and get
             * source_addr now. */
            if((token = strtok(NULL," ")) == NULL)
            {
                fprintf(stdout,"PIXILATE ACL PARSING ERROR: "
                    "\"host\" was not followd by source_addr.\n");
                pacl->t_errors++;
                continue;
            }      

            if((pacl->source_addr =
                libnet_name2addr4(l,token,LIBNET_DONT_RESOLVE)) == -1)
            {
                fprintf(stdout,"PIXILATE ADDRESS ERROR: "
                    "Invalid source address %s\n",token);
                pacl->t_errors++;
                continue;
            }
        }
        /* If not 'any' or 'host', process hostmask and netmask */
        else
        {
            if(strncmp(token,"127",3) ==0)
            {
                fprintf(stdout,
                    "PIXILATE NOTE: Libnet cannot send packets from %s\n",
                    token);
                continue;
            }
            else if(strncmp(token,"0.0.0.0",7) == 0) /* Same as 'any' */
            {
                if(pacl->def_source_addr == 0)
                {
                    /* Get random source address if default source address for
                     * 'any' was not supplied via -s command line option */
                    libnet_seed_prand(l);
                    pacl->source_addr = libnet_get_prand(LIBNET_PRu32);
                }
                else
                    pacl->source_addr = pacl->def_source_addr;
            }
            else
            {
                memset(&ip.s_addr, 0, 4);
                memset(&netmaskaddr.s_addr, 0,4);
                if((hostmask = inet_addr(token)) == INADDR_NONE)
                {
                    fprintf(stdout,"PIXILATE ADDRESS ERROR: "
                        "Invalid source address %s\n",token);
                    pacl->t_errors++;
                    continue;
                }
                else
                   ip.s_addr = hostmask;
        
            
                /* If we have a host mask, we *should* have a netmask. 
                 * We should only reach the NULL condition with a Simple ACL
                 * (0-99) That did not properly use the "host" identifier or 
                 * did not supply the netmask of "255.255.255.255" */
                if((token = strtok(NULL," ")) == NULL)
                    token = "255.255.255.255";
            
                if(pacl->options & ROUTER) /* Cisco IOS uses inverted masks */
                {
                    netmask = ~inet_addr(token);
                    /* NOTE: If a router acl (-r option) is parsed
                     * with a inverse mask of "0.0.0.0", it will become
                     * INADDR_NONE = "255.255.255.255" = 0xffffffff
                     * We catch this divide by 0 error condition below */
                    token = libnet_addr2name4(netmask,LIBNET_DONT_RESOLVE);
                }

                if((netmask = inet_addr(token)) == INADDR_NONE)
                {
                    fprintf(stdout,"PIXILATE ADDRESS ERROR: "
                        "Invalid netmask: %s\n",token);
                    pacl->t_errors++;
                    continue;
                }
               
                if(netmask == 0xffffffff) /* Divide by 0 catcher */
                {
                    pacl->destination_addr =
                       libnet_name2addr4(l,inet_ntoa(ip),LIBNET_DONT_RESOLVE);
                }
                else
                {
                    start = ip;
                    end.s_addr = ip.s_addr |= ~htonl(netmask);
          
                    i = ntohl(end.s_addr) - ntohl(start.s_addr);
                    random_int = ntohl(start.s_addr) + rand()%i; 
                    current.s_addr = htonl(random_int);
                    pacl->source_addr =
                        libnet_name2addr4(l,inet_ntoa(current),
                            LIBNET_DONT_RESOLVE);
                }
            }
        }
       
        if(aclid < 100)
        {
            /* send simple (0-99) ACLs to specific internal host
             * since they do not provide a destination address
             * skip the next section */ 
            pacl->destination_addr = pacl->def_destination_addr;
        }
        /* Only Extended ACLS (100-199) need the following parsing */
        else if((aclid > 99) && (aclid < 200))
        {
            if((token = strtok(NULL," ")) == NULL)
            {
                fprintf(stdout,"PIXILATE EXTENDED ACL PARSING ERROR: "
                    "Nothing found after source_addr. Should this be a "
                    "Simple ACL(0-99)\?\n");
                pacl->t_errors++;
                continue;
            }      
            /* Skip ACLs with object-group and obj_group_id */
            if(strncmp(token,"object-group",12) == 0)
            {
                fprintf(stdout,
                    "PIXILATE NOTE: Skipped object-group obj_grp_id ACL.\n");
                continue;
            }      
            /* If we are not processing an icmp rule, process source port */
            if(!(pacl->rule & ICMP_RULE))
            {
                if((strncmp(token,"lt",2) == 0)   ||
                    (strncmp(token,"gt",2) == 0)  ||
                    (strncmp(token,"eq",2) == 0)  ||
                    (strncmp(token,"neq",3) == 0) ||
                    (strncmp(token,"range",5) == 0))
                {
                    if(strncmp(token,"eq",2) == 0)
                    {
                        if((token = strtok(NULL," ")) == NULL)
                        {
                            fprintf(stdout, "PIXILATE ACL PARSING ERROR: "
                                "eq not followed by a source port.\n");
                            pacl->t_errors++;
                        } 
                        pacl->source_port = port_lookup(token,pacl);
                    }
                    else if(strncmp(token,"lt",2) == 0)
                    {
                        if((token = strtok(NULL," ")) == NULL)
                        {
                            fprintf(stdout, "PIXILATE ACL PARSING ERROR: "
                                "lt not followed by a source port.\n");
                            pacl->t_errors++;
                            continue;
                        }      
                        range = port_lookup(token,pacl) - 1;
                        pacl->source_port = 1 + rand()%range;
                    }
                    else if(strncmp(token,"gt",2) == 0)
                    {
                        if((token = strtok(NULL," ")) == NULL)
                        {
                            fprintf(stdout, "PIXILATE ACL PARSING ERROR: "
                                "gt not followed by a source port.\n");
                            pacl->t_errors++;
                            continue;
                        }      
                        start_port = port_lookup(token,pacl);
                        range = 65534 - start_port; /* 65534 isnt a typo */
                        pacl->source_port = start_port + rand()%range;
                    }
                    else if(strncmp(token,"neq",3) == 0)
                    {
                        if((token = strtok(NULL," ")) == NULL)
                        {
                            fprintf(stdout, "PIXILATE ACL PARSING ERROR: "
                                "neq not followed by a source port.\n");
                            pacl->t_errors++;
                            continue;
                        }      
                        range = 65534; /* 65534 isnt a typo */
                        pacl->source_port = rand()%range;
                        /* Make sure we are != */
                        if(pacl->source_port == port_lookup(token,pacl)) 
                            pacl->source_port = rand()%range;
                    }
                    else if(strncmp(token,"range",5) == 0)
                    {
                        if((token = strtok(NULL," ")) == NULL)
                        {
                            fprintf(stdout, "PIXILATE ACL PARSING ERROR: range "
                                "not followed by a starting source port.\n");
                            pacl->t_errors++;
                            continue;
                        }      
                        start_port = port_lookup(token,pacl);
                        if((token = strtok(NULL," ")) == NULL)
                        {
                            fprintf(stdout, "PIXILATE ACL PARSING ERROR: "
                                "range did not supply ending source port.\n");
                            pacl->t_errors++;
                            continue;
                        }      
                        end_port = port_lookup(token,pacl);
                        range = end_port - start_port;
                        pacl->source_port = start_port + rand()%range;
                    }
                    if((token = strtok(NULL, " ")) == NULL)
                    {
                        fprintf(stdout,"PIXILATE ACL PARSING ERROR: "
                            "Unexpected end of ACL.\n");
                        pacl->t_errors++;
                        continue;
                    }      
                }
            
                /* Skip ACLs with object-group and obj_group_id */
                if(strncmp(token,"object-group",12) == 0)
                {
                    fprintf(stdout, "PIXILATE NOTE: "
                            "Skipped object-group obj_grp_id ACL.\n");
                    continue;
                }      
            }

            /* Destination address and destination netmask for Extended
             * (100-100) ACLs. */
            if((strncmp(token,"any",3) == 0) ||  
                (strncmp(token,"0.0.0.0",7) == 0)) /* Same as 'any' */
                /* send 'any' rules to specific internal host */ 
                pacl->destination_addr = pacl->def_destination_addr;
            else if(strncmp(token,"host",4) == 0)
            {
                /* host is followed by dest_addr, so skip "host" and get
                 * dest_addr now. */
                if((token = strtok(NULL," ")) == NULL)
                {
                    fprintf(stdout,"PIXILATE ACL PARSING ERROR: "
                        "\"host\" was not followd by source_addr.\n");
                    pacl->t_errors++;
                    continue;
                }      
         
                if((pacl->destination_addr =
                    libnet_name2addr4(l,token,LIBNET_DONT_RESOLVE)) == -1)
                {
                    fprintf(stdout,"PIXILATE ADDRESS ERROR: "
                        "Invalid destination address %s\n",token);
                    pacl->t_errors++;
                    continue;
                }
            }
            /* If not "any" or "host", it is a destination address */
            else
            {
                memset(&ip.s_addr, 0, 4);
                memset(&netmaskaddr.s_addr, 0,4);

                if((hostmask = inet_addr(token)) == INADDR_NONE)
                {
                    fprintf(stdout,"PIXILATE ADDRESS ERROR: "
                        "Invalid source address %s\n",token);
                    pacl->t_errors++;
                    continue;
                }
                else
                    ip.s_addr = hostmask;

            /* If we have a host mask, we *should* have a netmask.
             * We should only reach the NULL condition with a Simple ACL (0-99)
             * That did not properly use the "host" identifier or 
             * did not supply the netmask of "255.255.255.255" */
            if((token = strtok(NULL," ")) == NULL)
                token = "255.255.255.255";

            if(pacl->options & ROUTER) /* Cisco IOS uses inverted masks */
            {
                netmask = ~inet_addr(token);
                /* NOTE: If a router acl (-r option) is parsed
                 * with a inverse mask of "0.0.0.0", it will become
                 * INADDR_NONE = "255.255.255.255" = 0xffffffff
                 * We catch this divide by 0 error condition below */
                token = libnet_addr2name4(netmask,LIBNET_DONT_RESOLVE);
            }
            /* NOTE: The same divide by 0 error condition will arise
             * on a PIX ACL if the netmask is 255.255.255.255 and it is
             * also caught below */
            
            if((netmask = inet_addr(token)) == INADDR_NONE)
            {
                fprintf(stdout,"PIXILATE ADDRESS ERROR: "
                    "Invalid netmask: %s\n",token);
                pacl->t_errors++;
                continue;
            }
            else
                netmask = ntohl(netmaskaddr.s_addr);
            
            if(netmask == 0xffffffff) /* Divide by 0 catcher */
            {
                pacl->destination_addr =
                libnet_name2addr4(l,inet_ntoa(ip),LIBNET_DONT_RESOLVE);
            }
            else
            {
                start = ip;
                end.s_addr = ip.s_addr |= ~htonl(netmask);
                i = ntohl(end.s_addr) - ntohl(start.s_addr);
                random_int = ntohl(start.s_addr) + rand()%i; 
                current.s_addr = htonl(random_int);
                pacl->destination_addr =
                    libnet_name2addr4(l,inet_ntoa(current),LIBNET_DONT_RESOLVE);
            }
         }
         
         /* Get icmp_type if applicable and send ICMP packets */
         if(pacl->rule & ICMP_RULE)
         {
            if((token = strtok(NULL, " ")) == NULL)
               token = "echo";

            pacl->icmp_type = cisco_icmp_type_lookup(token, pacl);
            packetize(pacl, l);
            /* NOTE ICMP Packets are now sent */
            continue; 
         }
         /* Pixilate is far enough along that if there is a strtok() error
          * we can safely ignore it */
         else if((token = strtok(NULL," ")) == NULL)
               token = "pixilate"; // junk value to stop seg fault
               
         /* Skip ACLs with object-group and obj_group_id */
         if(strncmp(token,"object-group",12) == 0)
         {
            fprintf(stdout,
               "PIXILATE NOTE: Skipped object-group obj_grp_id ACL.\n");
            continue;
         }

         if(strncmp(token,"log",3) == 0)
         {
            if((token = strtok(NULL, " ")) != NULL)
            {
               if(strncmp(token,"default",7) == 0)
                  fprintf(stdout,
                          "PIXILATE NOTE: Default ACL logging restored.\n");
               else if(strncmp(token,"disable",7) == 0)
                  fprintf(stdout,
                          "PIXILATE NOTE: ACL logging disabled.\n");
               else
                  fprintf(stdout,
                            "PIXILATE NOTE: ACL logging level is %s\n",token);
            }
         }   

         if(pacl->rule & IPSEC_RULE)
         {
            packetize(pacl, l);
            continue; 
         }  
      } /* End of specific Extended (100-199) parsing */

      if((pacl->rule & TCP_RULE) ||  (pacl->rule & UDP_RULE) ||
         (pacl->rule & IGMP_RULE) || (pacl->rule & IP_RULE))
      {
         /* Lookup operator for destination port */
         if(((strncmp(token,"lt",2) == 0)) ||
            ((strncmp(token,"gt",2) == 0)) ||
            ((strncmp(token,"eq",2) == 0)) ||
            ((strncmp(token,"neq",3) == 0)) ||
            ((strncmp(token,"range",5) == 0)))
         {
            if(strncmp(token,"eq",2) == 0)
            {
               if((token = strtok(NULL," ")) == NULL)
               {
                  fprintf(stdout, "PIXILATE ACL PARSING ERROR: "
                     "eq not followed by a destination port.\n");
                  pacl->t_errors++;
                  continue;
               }      
               pacl->destination_port = port_lookup(token,pacl);
            }
            else if(strncmp(token,"lt",2) == 0)
            {
               if((token = strtok(NULL," ")) == NULL)
               {
                  fprintf(stdout, "PIXILATE ACL PARSING ERROR: "
                     "lt not followed by a destination port.\n");
                  pacl->t_errors++;
                  continue;
               }      
               range = port_lookup(token,pacl) - 1;
               pacl->destination_port = 1 + rand()%range;
            }
            else if(strncmp(token,"gt",2) == 0)
            {
               if((token = strtok(NULL," ")) == NULL)
               {
                  fprintf(stdout, "PIXILATE ACL PARSING ERROR: "
                     "gt not followed by a destination port.\n");
                  pacl->t_errors++;
                  continue;
               }      
               start_port = port_lookup(token,pacl);
               range = 65534 - start_port; /* 65534 isnt a typo */
               pacl->destination_port = start_port + rand()%range;
            }
            else if(strncmp(token,"neq",3) == 0)
            {
               if((token = strtok(NULL," ")) == NULL)
               {
                  fprintf(stdout, "PIXILATE ACL PARSING ERROR: "
                     "eq not followed by a destination port.\n");
                  pacl->t_errors++;
                  continue;
               }      
               range = 65534; /* 65534 isnt a typo */
               pacl->destination_port = rand()%range;
               if(pacl->destination_port == port_lookup(token,pacl))
                  pacl->destination_port = rand()%range;
            }
            else if(strncmp(token,"range",5) == 0)
            {
               if((token = strtok(NULL," ")) == NULL)
               {
                  fprintf(stdout, "PIXILATE ACL PARSING ERROR: "
                     "range not followed by a starting destination port.\n");
                  pacl->t_errors++;
                  continue;
               }      
               start_port = port_lookup(token,pacl);
               if((token = strtok(NULL," ")) == NULL)
               {
                  fprintf(stdout, "PIXILATE ACL PARSING ERROR: "
                     "range did not have an ending destination port.\n");
                  pacl->t_errors++;
                  continue;
               }      
               end_port = port_lookup(token,pacl);
               range = end_port - start_port;
               pacl->destination_port = start_port + rand()%range;
            }
        }

        if((token = strtok(NULL," ")) != NULL)
        {
            if((strncmp(token,"payload",7)) == 0)
            {
                if((token = strtok(NULL,"\"")) != NULL)
                {
                    pacl->payload = token;
                    pacl->payload_s = strlen(token);
                }
            }
        }
        else
        {
            if(pacl->def_payload != NULL)
            {
                pacl->payload = pacl->def_payload;
                pacl->payload_s = strlen(pacl->def_payload);
            }
            else
            {
                pacl->payload = NULL;
                pacl->payload_s = 0;
            }
        } 
        packetize(pacl, l); 
        
      } /* end of is_tcp_rule || is_udp_rule */
      else
      {
         fprintf(stdout,
               "Access list entry appears to be unsupported at this time.\n");
         continue;
      }

   } /* End of while fgets() */
   fclose(input);
   fprintf(stdout,
      "____________________________________________________________________\n");
   fprintf(stdout,"Completed processing of %s\n",inputfile); 
   return(0);
}
