/*    pixilate.c
 *    Copyright 2002-2004 Kirby Kuehl (vacuum@users.sourceforge.net)
 *    This file is part of pixilate
 *    http://winfingerprint.sourceforge.net/pixilate.php
 *
 *    pixilate is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    pixilate is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *    You should have received a copy of the GNU General Public License
 *    along with pixilate; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "pixilate.h"

int main(int argc, char *argv[])
{
   libnet_t *l;
   int c        = 0,    /* Used by getopt */ 
     old        = 0;    /* used for dup(stdout) restoration */
   char errbuf[LIBNET_ERRBUF_SIZE];
   char inputfile[PATH_MAX];   /* File containing ACLs to test  */
   char outputfile[PATH_MAX];
   struct acl *pacl;
   FILE *output = NULL;

#ifdef _WIN32
   WSADATA wsaData;
   if((WSAStartup(0x0202, &wsaData )) != 0)
   {
		fprintf(stderr,"Unable to initialize Ws2_32.dll\n");
		exit(EXIT_FAILURE);
   }
#endif

   if((pacl = malloc(sizeof(struct acl))) == NULL)
   {
      fprintf(stderr,"Unable to allocate memory.\n");
      exit(EXIT_FAILURE);
   }
   
   /* Initialize pacl variables */ 
   pacl->protocol = 0;
   pacl->source_addr = 0;
   pacl->destination_addr = 0;
   pacl->def_destination_addr = 0;
   pacl->def_source_addr = 0;
   pacl->icmp_type = 0;
   pacl->gateway_addr = 0;
   pacl->id = 113;                  // Default IP ID if not specified 
   pacl->seq = 331;                 // Default SEQ NUM if not specified
   pacl->def_payload = NULL;        // Default pauload if not specified
   pacl->payload = NULL;            // Default payload if not specified
   pacl->payload_s = 0;             // Default payload size if not specified
   pacl->def_source_port = 1025;    // Default source port if not specified
   pacl->def_destination_port = 80; // Default desination port if not specified
   pacl->options = 0;           // Process permit and deny by default
   pacl->options |= BOTH;       // both by default
   memset(inputfile,0,PATH_MAX); 

   if(argc < 2)
      usage(argv[0]);

   while((c = getopt(argc,argv,"d:D:f:g:hi:m:o:rs:S:p:q:v"))!= -1)
   {
      switch(c)
      {
         case 'f': // file name of access list
            snprintf(inputfile,PATH_MAX-1,"%s",optarg);
            break;
         case 'h':
            usage(argv[0]);
            break;
         case 'i': // ip id
            pacl->id = atoi(optarg);
            break;
         case 'g':
            if((pacl->gateway_addr =
                  libnet_name2addr4(l,optarg,LIBNET_DONT_RESOLVE)) == -1)
            {
               fprintf(stderr,"Bad gateway IP address %s\n",optarg);
               exit(-1);
            }
            break;
         case 'm': // acl_type
            if((strncmp(optarg,"deny",4)) == 0)
            {
               pacl->options &= ~BOTH;
               pacl->options |= DENY_ONLY;
            }
            else if((strncmp(optarg,"permit",6)) == 0)
            {
               pacl->options &= ~BOTH;
               pacl->options |= PERMIT_ONLY;
            }
            else
               pacl->options |= BOTH; 
            break;
         case 'o':
            snprintf(outputfile,PATH_MAX-1,"%s",optarg);
            pacl->options |= USE_OUTPUT;
            break;
         case 'p': // payload
            pacl->def_payload = optarg;
            pacl->payload_s = strlen(optarg);
            break;
         case 'q': // sequence id
            pacl->seq = atoi(optarg);
            break;
         case 'r':
            pacl->options |= ROUTER;
            break;
         case 's': // default source address for 'any' rules
            if((pacl->def_source_addr =
                  libnet_name2addr4(l,optarg,LIBNET_DONT_RESOLVE)) == -1)
            {
               fprintf(stderr,"Bad default source IP address %s\n",optarg);
               exit(-1);
            }
            break;
         case 'S': // default source port for 'any' rules
            pacl->def_source_port = atoi(optarg); 
            break;
         case 'd': // default destination address for 'any' rules
            if((pacl->def_destination_addr =
                  libnet_name2addr4(l,optarg,LIBNET_DONT_RESOLVE)) == -1)
            {
               fprintf(stderr,"Bad default destination IP address %s\n",optarg);
               exit(-1);
            }
            break;
         case 'D': // default destination port for 'any' rules
            pacl->def_destination_port = atoi(optarg); 
            break;
         case 'v':
            pacl->options |= VERBOSE;
            break;
      }
   }
   
   if(inputfile[0] == '\0')
   {
      fprintf(stderr,
            "-f <PIX formatted access-list> is a required argument.\n"
            "See usage by typing %s -h\n",argv[0]);
      exit(-1);
   }
   
   if(pacl->def_destination_addr == 0)
   {
      fprintf(stderr,
            "-d <default destination ip> is a required argument.\n"
            "The default desination ip is needed so to ensure that the \n"
            "packet reaches the intended target when an Extended ACL\n"
            "with 'any' as a destination or a Simple ACL is encountered.\n"
            "See usage by typing %s -h\n",argv[0]);
      exit(-1);
   }

   if(pacl->options & USE_OUTPUT)
   {
     if((old = dup(1)) == -1)
     {
        fprintf(stderr,"Unable to dup() stdout.\n");
        exit(-1);
     }
     if((output = fopen(outputfile,"w")) == NULL)
     {
        fprintf(stderr,"Unable to open %s for writing.\n",outputfile);
        exit(-1);
     }
     if(dup2(fileno(output),1) == -1)
     {
        fprintf(stderr,"Unable to dup2 stdout.\n");
        exit(-1);
     }
   }
   
   /* Initialie Libnet */
   if((l = libnet_init(
                       LIBNET_RAW4,      /* Raw sockets using IPv4            */
                       NULL,             /* network interface                 */
                       errbuf            /* character buffer holding error    */
                       )) == NULL)       /* libnet_init returns NULL on error */
   {
      fprintf(stderr,"libnet_init(failed: %s\n", errbuf);
      exit(EXIT_FAILURE);
   }
   else
      fprintf(stdout,"Libnet initialized for interface: %s\n",
              libnet_getdevice(l));
  
   /* Parse inputfile */
   cisco_parser(inputfile, pacl,l);
   fflush(stdout);

   if(pacl->options & USE_OUTPUT)
   {
      fclose(output);
      dup2(old,1);
   }
  
   fprintf(stdout,"Total ACL errors          %d\n",pacl->t_errors);
   fprintf(stdout,"Total Permit Packets Sent %d\n",pacl->t_permits);
   fprintf(stdout,"Total Deny Packets Sent   %d\n",pacl->t_denies);
   fprintf(stdout,"Total Packets Sent        %d\n",
         pacl->t_permits + pacl->t_denies);

   if(pacl != NULL)
      free(pacl);
   libnet_destroy(l);                    /* Tear down all memory structures   */
   return(EXIT_SUCCESS); 
}

void usage(char *progname)
{
   fprintf(stdout,
   "ACL Validation Tool 0.4 by Kirby Kuehl (vacuum@users.sourceforge.net)\n");
   fprintf(stdout,"Usage: %s [-dDfghipqsSv]\n",progname);
   fprintf(stdout,
         "-d <Default Destination Address>\n");
   fprintf(stdout,
       "\t Destination Address to be used when destination address is 'any'\n");
   fprintf(stdout,"\t This MUST be an IP on the other side of the firewall.\n");
   fprintf(stdout,
         "-D <Default Destination Port>\n");
   fprintf(stdout,
         "\t Destination Port to be used when destination port is 'any'\n");
   fprintf(stdout,"\t If omitted, the default destination port is 80.\n");
   fprintf(stdout,"-f <PIX formatted access-list>\n");
   fprintf(stdout, "-g <gateway address>\n");
   fprintf(stdout,"\t Optional parameter to specify default gateway.\n");
   fprintf(stdout,
         "\t This parameter is only used for ICMP redirect packets.\n");
   fprintf(stdout,"-h\t Display usage.\n");
   fprintf(stdout,"-i <ID>\t Optional parameter to specify ip id.\n");
   fprintf(stdout,"\t If omitted, ID is set to 113\n");
   fprintf(stdout,"-m <permit|deny>\n");
   fprintf(stdout,"\t Process only permit or only deny ACLs.\n");
   fprintf(stdout,"\t If omitted, process both types.\n");  
   fprintf(stdout,"-o <output file>\n");
   fprintf(stdout,
         "\t Optional parameter redirecting standard output to filename.\n");
   fprintf(stdout,"-p <payload>\n");
   fprintf(stdout,"\t Optional parameter to specify payload.\n");
   fprintf(stdout,"-q <SEQ> Optional parameter to specify sequence number.\n");
   fprintf(stdout,"\t If omitted, SEQ is set to 331\n");
   fprintf(stdout,
         "-r\t Supplied ACL is in IOS format (inverse mask) rather than\n");
   fprintf(stdout,"\t the default PIX format (normal mask).\n");
   fprintf(stdout,
         "-s <Default Source Address>\n");
   fprintf(stdout,
         "\t Source Address to be used when source address is 'any'\n");
   fprintf(stdout,"\t If omitted, a random source address is chosen.\n");
   fprintf(stdout,"-S <Default Source Port>\n");
   fprintf(stdout,"\t Source Port to be used when source port is 'any'\n");
   fprintf(stdout,"\t If omitted, the default source port is 1025.\n");
   fprintf(stdout,"\t If omitted, the default destination port is 80.\n");
   fprintf(stdout,"-v \t Display statistics for each packet sent.\n");
   exit(EXIT_FAILURE);
}
